from time import sleep

import pytest

from pedros.timed import timed


def test_timed_sync():
    @timed
    def func():
        sleep(0.02)
        return True

    assert func()
    elapsed = getattr(func, "__last_elapsed__")
    assert isinstance(elapsed, float)
    assert elapsed >= 0.01


@pytest.mark.asyncio
async def test_timed_async():
    @timed
    async def func():
        sleep(0.02)
        return True

    assert await func()
    elapsed = getattr(func, "__last_elapsed__")
    assert isinstance(elapsed, float)
    assert elapsed >= 0.01

def test_timed_error():
    @timed
    def func():
        sleep(0.02)
        raise ValueError("Intended error raised in test")

    with pytest.raises(ValueError):
        func()

    elapsed = getattr(func, "__last_elapsed__")
    assert isinstance(elapsed, float)
    assert elapsed >= 0.01


from time import perf_counter
from typing import Callable, ParamSpec, TypeVar

from pedros.decorator_factory import CallContext, make_around_decorator
from pedros.logger import get_logger

logger = get_logger()

P = ParamSpec("P")
R = TypeVar("R")


def _before(ctx: CallContext[P, R]) -> None:
    ctx.extra["start_time"] = perf_counter()


def _after(ctx: CallContext[P, R]) -> None:
    start_time = ctx.extra.get("start_time")
    if start_time is None:
        return

    elapsed = perf_counter() - start_time

    setattr(ctx.wrapped, "__last_elapsed__", elapsed)
    logger.info(f"{ctx.wrapped.__name__} took {elapsed} seconds to execute.")


def timed(func: Callable[P, R]) -> Callable[P, R]:
    return make_around_decorator(before=_before, after=_after)(func)



import logging
from typing import Optional


def setup_logging(level: int = logging.INFO) -> None:
    """
    Configure the application's logging behavior.

    This function attempts to use Rich's ``RichHandler`` for enhanced,
    colorful, and trace-friendly logging. If Rich is not installed,
    it silently falls back to Python's standard logging configuration.
    See more about Rich (https://pypi.org/project/rich/).

    :param level: Logging level to use. Defaults to ``logging.INFO``.
    :type level: int
    :return: None
    """
    handler = None

    try:
        from rich.logging import RichHandler

        handler = RichHandler(rich_tracebacks=True)
        handlers = [handler]
        fmt = "%(message)s"
        datefmt = None

    except (ImportError, ModuleNotFoundError):
        handlers = None
        fmt = "%(asctime)s | %(levelname)-8s | %(message)s"
        datefmt = "%Y-%m-%d %H:%M:%S"

    logging.basicConfig(
        level=level,
        format=fmt,
        datefmt=datefmt,
        handlers=handlers,
    )

    logger = logging.getLogger(__name__)
    if handler is None:
        logger.debug("Rich library not found, using standard logging.")


def get_logger(name: Optional[str] = None) -> logging.Logger:
    """
    Return a logger instance.

    If no name is provided, the module's ``__name__`` is used.

    :param name: Name of the logger. If ``None``, defaults to the current module.
    :type name: str or None
    :return: A configured logger instance.
    """
    return logging.getLogger(name or __name__)



from __future__ import annotations

import inspect
from dataclasses import dataclass, field
from types import TracebackType
from typing import Any, Awaitable, Callable, Generic, ParamSpec, TypeVar, cast

import wrapt

P = ParamSpec("P")
R = TypeVar("R")


@dataclass
class CallContext(Generic[P, R]):
    wrapped: Callable[P, R]
    instance: Any | None
    args: P.args
    kwargs: P.kwargs
    result: R | Awaitable[R] | None = None
    error: BaseException | None = None
    extra: dict[str, Any] = field(default_factory=dict)

    def set_result(self, value: R | Awaitable[R]) -> None:
        self.result = value

    def set_error(self, err: BaseException) -> None:
        self.error = err


BeforeHook = Callable[[CallContext[P, R]], None]
AfterHook = Callable[[CallContext[P, R]], None]


class _HookContext(Generic[P, R]):
    def __init__(
        self,
        ctx: CallContext[P, R],
        before: BeforeHook[P, R] | None,
        after: AfterHook[P, R] | None,
    ) -> None:
        self._ctx = ctx
        self._before = before
        self._after = after

    def __enter__(self) -> None:
        if self._before is not None:
            self._before(self._ctx)
        return None

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> bool | None:
        if exc is not None:
            self._ctx.set_error(exc)

        if self._after is not None:
            self._after(self._ctx)

        return None


def make_around_decorator(
    *,
    before: BeforeHook[P, R] | None = None,
    after: AfterHook[P, R] | None = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:

    @wrapt.decorator
    def _wrapper(
        wrapped: Callable[P, R],
        instance: Any | None,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> R | Awaitable[R]:
        ctx: CallContext[P, R] = CallContext(
            wrapped=cast(Callable[P, R], wrapped),
            instance=instance,
            args=cast(P.args, args),
            kwargs=cast(P.kwargs, kwargs),
        )

        is_async = inspect.iscoroutinefunction(wrapped)

        if not is_async:
            with _HookContext(ctx, before, after):
                value = wrapped(*args, **kwargs)
                ctx.set_result(cast(R, value))
                return cast(R, value)

        async def _run_async() -> R:
            with _HookContext(ctx, before, after):
                wrapped_async = cast(Callable[..., Awaitable[R]], wrapped)
                value = await wrapped_async(*args, **kwargs)
                ctx.set_result(cast(R, value))
                return cast(R, value)

        coro = _run_async()
        ctx.set_result(coro)
        return coro

    return cast(Callable[[Callable[P, R]], Callable[P, R]], _wrapper)



from time import sleep

from pedros.logger import get_logger
from pedros.timed import timed


@timed
def long_func():
    sleep(0.1)


def main():
    logger = get_logger()
    logger.info("Hello World!")
    long_func()


if __name__ == "__main__":
    main()
